from validateit.utils import check_type, kwargs_matcher
from validateit.descriptors import FieldDescriptor
from contextlib import suppress
import typing


class BaseValidator(FieldDescriptor):
    def __set__(self, obj, val):
        super(BaseValidator, self).__set__(obj, val)

    def validate(self):
        ...


class EmptyValidator(BaseValidator):
    def __set__(self, obj, val):
        self.validate(obj, val)
        super(EmptyValidator, self).__set__(obj, val)

    def validate(self, obj, val):
        ...


class TypeValidator(BaseValidator):
    def __set__(self, obj, val):
        self.validate(obj, val)
        super(TypeValidator, self).__set__(obj, val)

    def validate(self, obj, val, strict_level=1):
        with suppress(KeyError):
            type_hint = obj.__annotations__[self.name]
            if not check_type(type_hint, val):
                raise TypeError(
                    "Unexpected type for '{}' (expected {} but found {})".format(
                        self.name, type_hint, type(val)
                    )
                )


class IntegerValidator(BaseValidator):
    def __set__(self, obj, val):
        self.validate(obj, val)
        super(IntegerValidator, self).__set__(obj, val)

    def validate(self, obj, val):
        if not check_type(int, val):
            raise TypeError(
                "Unexpected type for {} (expected int but found {})".format(
                    self.name, type(val)
                )
            )

        args = kwargs_matcher(
            self.kwargs, dict.fromkeys(("min", "max", "min_digits", "max_digits"))
        )

        if not args["min_digits"]:
            pass
        elif len(str(val)) >= args["min_digits"]:
            pass
        else:
            raise ValueError(
                "Integer doesn't capable of having minimum {} digits.".format(
                    args["min_digits"]
                )
            )

        if not args["max_digits"]:
            pass
        elif len(str(val)) <= args["max_digits"]:
            pass
        else:
            raise ValueError(
                "Integer doesn't capable of having maximum {} digits.".format(
                    args["max_digits"]
                )
            )

        if not args["max"]:
            pass
        elif val <= args["max"]:
            pass
        else:
            raise ValueError(
                "Integer is bigger than {} (maximum limit)".format(args["max"])
            )

        if not args["min"]:
            pass
        elif val >= args["min"]:
            pass
        else:
            raise ValueError(
                "Integer is not bigger than {} (minimum limit)".format(args["min"])
            )


class StringValidator(BaseValidator):
    def __set__(self, obj, val):
        self.validate(obj, val)
        super(StringValidator, self).__set__(obj, val)

    def validate(self, obj, val):
        if not check_type(str, val):
            raise TypeError(
                "Unexpected type for {} (expected str but found {})".format(
                    self.name, type(val)
                )
            )

        args = kwargs_matcher(
            self.kwargs,
            dict.fromkeys(("min_chars", "max_chars", "startswith", "endswith")),
        )

        if not args["min_chars"]:
            pass
        elif len(val) >= args["min_chars"]:
            pass
        else:
            raise ValueError(
                "String doesn't capable of having minimum {} chars.".format(
                    args["min_chars"]
                )
            )

        if not args["max_chars"]:
            pass
        elif len(val) <= args["max_chars"]:
            pass
        else:
            raise ValueError(
                "String doesn't capable of having maximum {} chars.".format(
                    args["max_chars"]
                )
            )

        if not args["startswith"]:
            pass
        elif val.startswith(args["startswith"]):
            pass
        else:
            raise ValueError("String isn't starts with {}".format(args["startswith"]))

        if not args["endswith"]:
            pass
        elif val.endswith(args["endswith"]):
            pass
        else:
            raise ValueError("String isn't ends with {}".format(args["endswith"]))
